#!/usr/bin/env python3.6
# -*- coding: utf-8 -*-
import glob
import os
import warnings

from lxml import etree as ElementTree
from yaml import load, dump

from .components import warn_if_missing_templates, TemplateDefinition

try:
    from yaml import CLoader as Loader, CDumper as Dumper
except ImportError:
    from yaml import Loader, Dumper


class Transpiler(object):
    template_cls = TemplateDefinition

    def __init__(self, name: str = "Template Group"):
        self.name = name
        self.templates = {}

    def import_from_yml(self, path: str):
        """
        :param path: to yml file
        :return:
        """

        if not os.path.isfile(path):
            warnings.warn('YML file not found: ' + path + '. Skipping')
            return

        # I had a reason for not using `with`...
        # I don't remember what it was though :)
        stream = open(path)
        data = load(stream, Loader=Loader)
        stream.close()

        if data is None:
            warnings.warn('YML File is empty or has invalid data:'
                          + path + '. Skipping')
            return

        templates = [self.template_cls.build_from_yml(x) for x in data]

        templates.sort(key=lambda x: x.name.lower())

        templates = {template.name: template for template in templates}

        self.templates.update(templates)

        return self

    def import_from_sublimetext(self, file_or_dir_path: str) -> 'Transpiler':
        """
        :param file_or_dir_path: either to a specific file, or a directory
        with multiple
        snippets
        """

        if os.path.isdir(file_or_dir_path):
            file_or_dir_path = os.path.join(file_or_dir_path,
                                            '**',
                                            '*.sublime-snippet')

        files = glob.glob(file_or_dir_path, recursive=True)

        xml = [ElementTree.parse(file).getroot() for file in files]

        templates = [self.template_cls.build_from_snippet(child)
                     for child in xml]

        templates.sort(key=lambda x: x.name.lower())
        templates = {template.name: template for template in templates}

        self.templates.update(templates)

        return self

    def import_from_jetbrains(self, file_path: str) -> 'Transpiler':
        """
        :param file_path:
        """

        xml = ElementTree.parse(file_path).getroot()
        self.name = xml.attrib['group']

        templates = [self.template_cls.build_from_xml(child)
                     for child in xml]
        templates.sort(key=lambda x: x.name.lower())
        templates = {template.name: template for template in templates}

        self.templates.update(templates)

        return self

    # @warn_if_missing_templates
    def export_to_jetbrains(self, output_path: str) -> 'Transpiler':
        template_set = ElementTree.Element('templateSet')
        template_set.attrib['group'] = self.name

        for key, template in self.templates.items():
            xml_template = ElementTree.SubElement(
                template_set,
                'template',
                template.to_jetbrains_dict())

            for key, var in template.variables.items():
                ElementTree.SubElement(
                    xml_template,
                    'variable',
                    var.to_jetbrains_dict())

            context = ElementTree.SubElement(xml_template, 'context')

            for option in template.context_options:
                ElementTree.SubElement(
                    context,
                    'option',
                    option.to_jetbrains_dict())

        tree = ElementTree.ElementTree(template_set)

        tree.write(output_path, pretty_print=True)

        return self

    @warn_if_missing_templates
    def export_to_yml(self, output_path: str) -> 'Transpiler':
        stream = open(output_path, 'w')
        tpls = [tpl.to_yml_dict() for tpl in self.templates.values()]
        dump(tpls, stream, Dumper=Dumper, default_flow_style=False)
        stream.close()

        return self



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import os
import sys
from unittest import TestCase
from xml.etree import ElementTree

from shortbus import Transpiler
from shortbus.components import VariableDefinition, TemplateDefinition, \
    ContextDefinition
from shortbus.constants import CONTEXT

try:
    from yaml import CLoader as Loader, CDumper as Dumper, dump
except ImportError:
    from yaml import Loader, Dumper


class GeneratorTest(TestCase):
    def setUp(self):
        self.directory = os.path.dirname(os.path.realpath(__file__))
        self.valid_yml_path = os.path.join(
            self.directory, 'data', 'valid.yml')
        self.warning_yml_path = os.path.join(
            self.directory, 'data', 'warns.yml')
        self.jetbrains_xml = os.path.join(
            self.directory, 'data', 'jetbrains.xml')

        self.data_dir = os.path.join(self.directory, 'data')

        self.generator = Transpiler('Djaneiro: Models') \
            .import_from_yml(self.valid_yml_path)

        self.yml_templates = Transpiler().import_from_yml(
            self.valid_yml_path).templates

        self.xml_templates = Transpiler().import_from_jetbrains(
            self.jetbrains_xml).templates

    def tearDown(self):
        pass

    def test_default_values_for_template_and_raw(self):
        variables = {
            'var': VariableDefinition('var'),
            'end': VariableDefinition('end')
        }

        value = '$var$ = models.AutoField($var$)$end$'
        default = TemplateDefinition(
            name='raw',
            value=value,
            variables=variables,
        )

        template = self.yml_templates['raw']

        self.assertEqual(default, template)

        value = '$moo$\n$joo$'

        variables = {
            'moo': VariableDefinition('moo'),
            'joo': VariableDefinition('joo')
        }

        default.value = value
        default.variables = variables
        default.name = 'template'

        template = self.yml_templates['template']

        self.assertEqual(default, template)

    def test_variable_default_definitions(self):
        value = '$weep$ $graa:nah$ $weep:$ $:ni$ $ni$ $$'

        template = self.yml_templates['bah']
        variables = template.variables

        print(variables)

        self.assertEqual(variables['weep'].defaultValue, 'crabrabbit')
        self.assertEqual(variables['graa'].defaultValue, 'hard')
        self.assertEqual(variables['VAR4'].defaultValue, 'ni')
        self.assertEqual(variables['ni'].defaultValue, '')

        self.assertEqual(template.value,
                         '$weep$ $graa$ $weep$ $VAR4$ $ni$ $VAR6$')
        self.assertEqual(len(variables), 5)

    def test_variable_typo_throws_warning(self):
        with self.assertWarns(Warning):
            Transpiler().import_from_yml(self.warning_yml_path)

    def test_create_from_jetbrains_format(self):
        tpl = self.xml_templates['mauto']

        variables = {
            'VAR': VariableDefinition(name='VAR', defaultValue='FIELDNAME'),
            'VAR2': VariableDefinition(name='VAR2')
        }

        template = TemplateDefinition(
            name='mauto',
            value='$VAR$ = models.AutoField($VAR2$)',
            toReformat=False,
            toShortenFQNames=True,
            variables=variables
        )

        self.assertEqual(tpl.to_jetbrains_dict(), template.to_jetbrains_dict())

    def test_jetbrains_export(self):
        group = 'Djaneiro: Models'
        a = TemplateDefinition(
            name='mauto',
            value='$VAR$ = models.AutoField($VAR2$)',
            toReformat=False,
            toShortenFQNames=True,
            variables={
                'VAR': VariableDefinition('VAR', 'FIELDNAME'),
                'VAR2': VariableDefinition('VAR2')
            }
        )

        generator = Transpiler(group)
        generator.templates = {'mauto': a}

        path = os.path.join(self.data_dir, 'bullshit.xml')
        generator.export_to_jetbrains(path)

        a = ElementTree.parse(path).getroot()
        b = ElementTree.parse(self.jetbrains_xml).getroot()

        self.assertEqual(a.tag, b.tag)
        template_a, template_b = a.findall('template')[0], \
                                 b.findall('template')[0]
        self.assertEqual(sorted(template_a.items()),
                         sorted(template_b.items()))

        variables_a = template_a.findall('variable')
        variables_b = template_b.findall('variable')
        variables = zip(variables_a, variables_b)

        for a, b in variables:
            self.assertEqual(sorted(a.items()),
                             sorted(b.items()))

        options_a = template_a.find('context').findall('option')
        options_b = template_b.find('context').findall('option')

        options = zip(options_a, options_b)
        for a, b in options:
            self.assertEqual(sorted(a.items()),
                             sorted(b.items()))

    def test_import_from_sublime(self):
        a = Transpiler('test')
        path = os.path.join(self.data_dir)
        sublime = a.import_from_sublimetext(self.data_dir)

        path = os.path.join(self.data_dir, 'jetbrains.xml')
        jetbrains = a.import_from_jetbrains(path)

        self.assertEqual(jetbrains, sublime)

    def test_custom_transpiler(self):
        class CppContext(ContextDefinition):
            default_name = CONTEXT.cpp.all

        class CppTemplate(TemplateDefinition):
            default_context = [CppContext()]

        class CppTranspiler(Transpiler):
            template_cls = CppTemplate

        transpiler = CppTranspiler('cpp')

        path = os.path.join(self.data_dir, 'cpp.yml')
        transpiler.import_from_yml(path)
        result = transpiler.templates['bah'].context_options[0]
        self.assertEqual(result.name, 'cpp')



#!/usr/bin/env python3.6
# -*- coding: utf-8 -*-
import re
import warnings
from collections import OrderedDict, Callable
from functools import wraps
from typing import Tuple, List
from xml import etree

from .constants import CONTEXT


def warn_if_missing_templates(func):
    @wraps(func)
    def method(self, *method_args, **method_kwargs):
        if self.templates == {}:
            raise KeyError(
                "You seem to be missing some templates.  If you've already "
                "imported templates,"
                "then you need to merge them `merge_all_templates()`")

        func(self, *method_args, **method_kwargs)

    return method


class ContextDefinition(object):
    default_name = CONTEXT.python.all

    def __init__(self, name: str = '', value: bool = True):
        self.name: str = name if name is not '' else self.default_name
        self.value: bool = value

    def __eq__(self, other):
        return self.name == other.name and self.value == other.value

    def to_jetbrains_dict(self):
        return {
            'name': self.name,
            'value': str(self.value).lower()
        }


class VariableDefinition(object):
    def __init__(self, name: str, defaultValue: str = '', expression: str = '',
                 alwaysStopAt: bool = True):
        self.name: str = name
        self.defaultValue: str = defaultValue.strip('"\'')
        self.expression: str = expression
        self.alwaysStopAt: bool = alwaysStopAt

    def __eq__(self, other):
        return self.name == other.name \
               and self.expression == other.expression \
               and self.defaultValue == other.defaultValue \
               and self.alwaysStopAt is other.alwaysStopAt

    def to_jetbrains_dict(self):
        return {
            'name': self.name,
            'expression': self.expression,
            'defaultValue': f'"{self.defaultValue}"',

            'alwaysStopAt': str(self.alwaysStopAt).lower()
        }


class TemplateDefinition(object):
    variable_prefix: str = 'VAR{}'
    VARIABLE_WRAPPER: str = '${}$'
    VARIABLE_REGEX = r'(?P<raw>\$(?P<variable>[\w_\.]*)?:?' \
                     r'(?P<default>[^\$]*)?\$)'
    SUBLIME_CONTENT_REGEX = r'\<\!\[CDATA\[(.*)\]\]'
    SUBLIME_VARIABLE_REGEX = r'(?P<raw>\$\{?(?P<variable>\d):?' \
                             r'(?P<default>[\w_\s]*)\}?)'
    default_context = [ContextDefinition()]

    def __init__(self,
                 name: str,
                 value: str = '',
                 variables: {} = None,
                 context_options: List[ContextDefinition] = None,
                 shortcut: str = 'TAB',
                 description: str = '',
                 toReformat: bool = True,
                 toShortenFQNames: bool = True):

        self.name = name
        self.value = value
        self.shortcut = shortcut
        self.description = description

        self.toReformat = toReformat
        self.toShortenFQNames = toShortenFQNames

        self.variables = variables or {}
        self.context_options = context_options or self.default_context

    def __eq__(self, other):
        equals = [
            'name',
            'value',
            'variables',
            'context_options',
            'shortcut',
            'description'
        ]

        for item in equals:
            if getattr(self, item) != getattr(other, item):
                return False

        same = ['toReformat', 'toShortenFQNames']

        for item in same:
            if getattr(self, item) is not getattr(other, item):
                return False

        return True

    def format_description_lazily(self):
        name = self.value.split('.')[0].split('(')[0]
        self.description = f"{name} ({self.value})"

    def to_jetbrains_dict(self):
        return {
            'name': self.name,
            'value': self.value,
            'toReformat': str(self.toReformat).lower(),
            'toShortenFQNames': str(self.toShortenFQNames).lower()
        }

    def to_yml_dict(self):
        return {
            'name': self.name,
            'raw': self.value,
            'toReformat': self.toReformat,
            'toShortenFQNames': self.toShortenFQNames,
            'variables': [var.__dict__ for var in self.variables.values()],
            'context': [options.__dict__ for options in self.context_options],
            'shortcut': self.shortcut,
            'description': self.description
        }

    @classmethod
    def build_from_yml(cls, yml: {}):
        """
        :param yml:
        :return:
        """
        try:
            name = yml.pop('name')
        except KeyError:
            msg = "your template needs a name. you should fix that, corndog"
            raise KeyError(msg)

        value = yml.pop('raw', None)

        if value is None:
            try:
                path = yml.pop('template')
                with open(path) as file:
                    value = file.read()

            except KeyError as _:
                msg = "Your stupid yml file is broke.  entry needs either a " \
                      "raw or template field. fix it!\n"
                raise KeyError(msg)
            except FileNotFoundError as _:
                raise FileNotFoundError(f'Aint no file. get out: {path}')

        context = yml.pop('context', [])
        context_options = [ContextDefinition(**x) for x in context]

        value, variables = parse_and_extract_variables(
            value, TemplateDefinition.VARIABLE_REGEX)

        for i, variable in enumerate(yml.pop('variables', [])):
            variable_name = variable.get('name')

            if variable_name not in variables.keys():
                message = "Variable '{variable_name}' not found in " \
                          "'{name}': {value}. Skipping..."
                warnings.warn(message)
                continue

            variables[variable_name] = VariableDefinition(**variable)

        return cls(
            name=name,
            value=value,
            context_options=context_options,
            variables=variables,
            **yml)

    @classmethod
    def build_from_xml(cls, xml: {}):
        attrs = xml.attrib
        name = attrs['name']
        raw_vars = xml.findall('variable')

        toReformat = attrs['toReformat'].lower()
        toShortenFQNames = attrs['toShortenFQNames'].lower()

        boolean_converter = {
            'true': True,
            'false': False
        }

        toReformat = boolean_converter[toReformat]
        toShortenFQNames = boolean_converter[toShortenFQNames]

        variables = [VariableDefinition(**var.attrib) for var in raw_vars]

        variables = {var.name: var for var in variables}

        context = xml.find('context')
        options = []

        for option in context.findall('option'):
            option_name = option.attrib['name']
            value = option.attrib['value'].lower()
            value = boolean_converter[value]

            options.append(
                ContextDefinition(name=option_name, value=value))

        return cls(
            name,
            value=attrs['value'],
            toReformat=toReformat,
            toShortenFQNames=toShortenFQNames,
            variables=variables,
            context_options=options
        )

    @classmethod
    def build_from_snippet(cls, xml: etree):
        name = xml.findtext('tabTrigger')
        content = xml.findtext('content')

        value, variables = parse_and_extract_variables(
            content, TemplateDefinition.SUBLIME_VARIABLE_REGEX)

        scope = xml.findtext('scope')
        context = ContextDefinition(scope.split('.')[-1].capitalize())

        description = xml.findtext('description')

        return cls(
            name=name,
            value=value,
            description=description,
            variables=variables,
            context_options=[context])


def parse_and_extract_variables(string: str,
                                regex: str,
                                variables=None) -> Tuple[str, OrderedDict]:
    pattern: re._pattern_type = re.compile(regex)
    string: str = string.replace("$:$", "$$")

    format_variable = TemplateDefinition.variable_prefix.format

    wrap_variable: Callable = TemplateDefinition.VARIABLE_WRAPPER.format

    matches = pattern.findall(string)
    variables = variables or OrderedDict()

    for i, match in enumerate(matches):
        raw_match, variable_name, default_value = match

        max_replace = 1 if variable_name == '' else -1

        if variable_name == '':
            variable_name = format_variable(i + 1)
        elif variable_name.isdigit():
            var = int(variable_name)
            variable_name = format_variable(var)

        variable_dict = VariableDefinition(
            name=variable_name,
            defaultValue=default_value)

        variables[variable_name] = variable_dict
        variable_name = wrap_variable(variable_name)
        string = string.replace(raw_match, variable_name, max_replace)

    return string, variables



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from setuptools import setup

with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read()

requirements = [
    'pyyaml',
    'lxml'
]

test_requirements = []

setup(
    name='shortbus',
    version='0.2.1',
    description="Tools to convert Sublime Text snippets into Jetbrains live templates and vice versa",
    long_description=readme + '\n\n' + history,
    author="Brian McClure",
    author_email='brian@mcclure.pw',
    url='https://github.com/brmc/shortbus',
    packages=[
        'shortbus',
    ],
    package_dir={'shortbus':
                 'shortbus'},
    include_package_data=True,
    install_requires=requirements,
    license="MIT license",
    zip_safe=False,
    keywords='shortbus',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3.6',
    ],
    test_suite='tests',
    tests_require=test_requirements
)



class Css:
    all = 'CSS',
    declaration = 'CSS_DECLARATION_BLOCK',
    property = 'CSS_PROPERTY_VALUE',
    ruleset = 'CSS_RULESET_LIST'


class Dart:
    all = 'DART'
    statement = 'DART_STATEMENT'
    top_level = 'DART_TOPLEVEL'


class Groovy:
    all = 'GROOVY'
    declaration = 'GROOVY_DECLARATION'
    expression = 'GROOVY_EXPRESSION'
    statement = 'GROOVY_STATEMENT'


class Html:
    all = 'HTML'
    text = 'HTML_TEXT'


class Java:
    all = 'JAVA_CODE'
    comment = 'JAVA_COMMENT'
    declaration = 'JAVA_DECLARATION'
    expression = 'JAVA_EXPRESSION'
    statement = 'JAVA_STATEMENT'
    string = 'JAVA_STRING'


class Js:
    all = 'JAVA_SCRIPT'
    jsx = 'JSX_HTML'
    expression = 'JS_EXPRESSION'
    statement = 'JS_STATEMENT'


class Kotlin:
    cls = 'KOTLIN_CLASS'
    comment = 'KOTLIN_COMMENT'
    expression = 'KOTLIN_EXPRESSION'
    obj_declaration = 'KOTLIN_OBJECT_DECLARATION'
    statement = 'KOTLIN_STATEMENT'
    top_level = 'KOTLIN_TOPLEVEL'


class C:
    all = 'c'
    declaration = 'OC_DECLARATION_C'
    expression = 'OC_EXPRESSION_C'
    statemnet = 'OC_STATEMENT_C'


class Cpp:
    all = 'cpp'
    declaration = 'OC_DECLARATION_CPP'
    expression = 'OC_EXPRESSION_CPP'
    statement = 'OC_STATEMENT_CPP'


class Php:
    all = 'PHP'
    class_member = 'PHP Class Member'
    comment = 'PHP Comment'
    expression = 'PHP Expression'
    statement = 'PHP Statement'
    string = 'PHP String Literal'


class Python:
    all = 'Python'
    django = 'Django'
    cls = 'Python_Class'


class Vue:
    all = 'Vue'
    component = 'VUE_COMPONENT_DESCRIPTOR'
    tag = 'VUE_INSIDE_TAG'
    script = 'VUE_SCRIPT'
    template = 'VUE_TEMPLATE'
    top_level = 'VUE_TOP_LEVEL'


class Xml:
    all = 'XML'
    text = 'XML_TEXT'


class ContextContainer:
    all = 'OTHER'
    xsl = 'XSL_TEXT'
    cucumber = 'CUCUMBER_FEATURE_FILE'
    coffee_script = 'CoffeeScript'
    haml = 'HAML'
    json = 'JSON'
    sql = 'SQL'
    twig = 'Twig'
    type_script = 'TypeScript'
    completion = 'COMPLETION'

    c = C
    css = Css
    cpp = Cpp
    dart = Dart
    groovy = Groovy
    html = Html
    java = Java
    js = Js
    kotlin = Kotlin
    vue = Vue
    php = Php
    python = Python
    xml = Xml


CONTEXT = ContextContainer
